from torch.utils.data import DataLoader, Dataset
import matplotlib.pyplot as plt
import pandas as pd
import random
import torch
from torch.autograd import Variable
import torch.nn as nn
from torch import optim
import torch.nn.functional as F
from scipy import spatial
# import tensorflow_hub as hub
import torch.nn.utils.rnn as rnn_utils


def str2vec(s=('abc', 'bcd')):
    if Config.embedding_type:
        emd = pd.read_csv("char-embedding-normal.csv")
    else:
        emd = pd.read_csv("char-embedding.csv")
    embeddings_mix = []
    for idx, str in enumerate(s):
        embeddings = []
        for j in str.lower():
            embeddings.append(emd[j].to_list())
        embeddings_mix.append(torch.tensor(embeddings))

    # batch not in first dimension
    e_mix = rnn_utils.pad_sequence(embeddings_mix, batch_first=False, padding_value=0.0)

    e1 = e_mix[:, :Config.train_batch_size]
    e2 = e_mix[:, Config.train_batch_size:]

    return e1, e2


# Returns cosine similarity of two vectors
# def cos_sim(a, b):
#     return 1 - spatial.distance.cosine(a, b)


class MyDataset(Dataset):
    def __init__(self, csv_file):
        self.csv_data = pd.read_csv(csv_file)

    def __getitem__(self, idx):
        idx1 = random.randint(0, self.csv_data.shape[0] - 1)
        # we need to make sure approx 50% of samples are in the same class
        should_get_same_class = random.randint(0, 1)
        if should_get_same_class:
            text1 = self.csv_data['concept1'][idx1]
            text2 = self.csv_data['concept2'][idx1]
            label = 1
        else:
            idx2 = random.randint(0, self.csv_data.shape[0] - 1)
            while True:
                if idx1 != idx2:
                    text1 = self.csv_data['concept1'][idx1]
                    text2 = self.csv_data['concept2'][idx2]
                    label = 0
                    break

        return text1, text2, label

    def __len__(self):
        return len(self.csv_data) * 2


class SiameseBLSTM(nn.Module):
    def __init__(self):
        super(SiameseBLSTM, self).__init__()
        self.hidden_size = 128
        self.embed_dim = Config.embed_dim
        self.bidirectional = True
        self.layer_size = 1
        self.lstm = nn.LSTM(input_size=self.embed_dim,
                            hidden_size=self.hidden_size,
                            num_layers=self.layer_size,
                            bidirectional=self.bidirectional)
        # [128*2] => [1]
        self.fc = nn.Linear(self.hidden_size * 2, 1)

    def forward_once(self, x):
        output, (hn, cn) = self.lstm(x)
        output = output.permute(1, 0, 2)  # [batch, seq_len, hidden_dim*2]
        output = self.fc(output)
        return output

    def forward(self, input1, input2):
        output1 = self.forward_once(input1)
        output2 = self.forward_once(input2)
        return output1, output2


class SiameseAttBLSTM(nn.Module):
    def __init__(self):
        super(SiameseAttBLSTM, self).__init__()
        self.hidden_size = 64
        self.embed_dim = Config.embed_dim
        self.bidirectional = True
        self.layer_size = 1

        self.lstm = nn.LSTM(input_size=self.embed_dim,
                            hidden_size=self.hidden_size,
                            num_layers=self.layer_size,
                            bidirectional=self.bidirectional)
        # [128*2] => [1]
        self.fc = nn.Linear(self.hidden_size * 2, 1)

        self.w_omega = nn.Parameter(torch.Tensor(self.hidden_size * 2, self.hidden_size * 2))
        self.u_omega = nn.Parameter(torch.Tensor(self.hidden_size * 2, 1))
        nn.init.uniform_(self.w_omega, -0.1, 0.1)
        nn.init.uniform_(self.u_omega, -0.1, 0.1)

    def attention_net(self, x):  # x:[batch, seq_len, hidden_dim*2]

        u = torch.tanh(torch.matmul(x, self.w_omega))  # [batch, seq_len, hidden_dim*2]
        att = torch.matmul(u, self.u_omega)  # [batch, seq_len, 1]
        att_score = F.softmax(att, dim=1)
        scored_x = x * att_score  # [batch, seq_len, hidden_dim*2]
        # context = torch.sum(scored_x, dim=1)
        #
        # return context
        return scored_x

    def forward_once(self, x):
        output, (hn, cn) = self.lstm(x)
        output = output.permute(1, 0, 2)  # [batch, seq_len, hidden_dim*2]
        attn_output = self.attention_net(output)
        output = self.fc(attn_output)
        return output

    def forward(self, input1, input2):
        output1 = self.forward_once(input1)
        output2 = self.forward_once(input2)
        return output1, output2


class ContrastiveLoss(torch.nn.Module):

    def __init__(self, margin=1.0):
        super(ContrastiveLoss, self).__init__()
        self.margin = margin

    def forward(self, output1, output2, label):
        euclidean_distance = F.pairwise_distance(output1, output2).squeeze()
        loss_contrastive = torch.mean(label * torch.pow(euclidean_distance, 2) +
                                      (1 - label) * torch.pow(torch.clamp(self.margin - euclidean_distance, min=0.0),
                                                              2))

        return loss_contrastive


class ContrastiveLossNormalized(torch.nn.Module):

    def __init__(self, margin=1.0):
        super(ContrastiveLossNormalized, self).__init__()
        self.margin = margin

    def forward(self, output1, output2, label):
        # numerator = torch.sqrt(torch.sum(torch.pow(torch.sub(output1, output2), 2), dim=1))
        # denominator = torch.add(torch.sqrt(torch.sum(torch.pow(output1, 2), dim=1)),
        #                         torch.sqrt(torch.sum(torch.pow(output2, 2), dim=1)))
        numerator = F.pairwise_distance(output1, output2).squeeze()
        denominator = torch.add(torch.norm(output1, dim=1), torch.norm(output2, dim=1)).squeeze()
        normalized_euclidean_distance = torch.div(numerator, denominator).reshape(-1)
        # print(numerator, "\n", denominator, '\n', normalized_euclidean_distance)
        item1 = (label * normalized_euclidean_distance)
        item2 = (1 - label) * torch.clamp(self.margin - normalized_euclidean_distance, min=0.0)
        loss_contrastive = torch.div(torch.mean(torch.add(item1, item2)), 2)

        return loss_contrastive


def text2vec2():
    with open("char-embeddings.txt") as f:
        text = f.read()


# def text2vec(text):
#     # string2char
#     text = [[j for j in i] for i in text]
#     pd_list = []
#     # input(seq_len, batch, input_size)
#     for i in text:
#         pd_list.append(torch.from_numpy(embed(i).numpy()))
#
#     # batch not in first dimension
#     text_mix = rnn_utils.pad_sequence(pd_list, batch_first=False, padding_value=0.0)
#
#     text1 = text_mix[:, :Config.train_batch_size]
#     text2 = text_mix[:, Config.train_batch_size:]
#
#     return text1, text2


class Config:
    train_batch_size = 4
    train_number_epochs = 4
    embed_dim = 300
    embedding_type = 1  # 1:normalized


if __name__ == "__main__":
    # embed = hub.load(r"..\universal-sentence-encoder_4")
    siamese_dataset = MyDataset("train.csv")
    train_dataloader = DataLoader(siamese_dataset,
                                  shuffle=True,
                                  num_workers=0,
                                  batch_size=Config.train_batch_size)
    # net = SiameseBLSTM()
    # net = SiameseBLSTM().cuda()
    net = SiameseAttBLSTM().cuda()
    # criterion = ContrastiveLoss()
    criterion = ContrastiveLossNormalized()
    optimizer = optim.Adam(net.parameters(), lr=0.001)

    counter = []
    loss_history = []
    iteration_number = 0
    for epoch in range(Config.train_number_epochs):
        for i, data in enumerate(train_dataloader):
            print(i, data)
            text1, text2, label = data
            text_mix = text1 + text2
            text1, text2 = str2vec(text_mix)
            # text1, text2 = text2vec(text_mix)
            # print(text1.shape, text2.shape)
            text1, text2, label = Variable(text1).cuda(), Variable(text2).cuda(), Variable(label).cuda()
            output1, output2 = net(text1, text2)
            optimizer.zero_grad()
            loss_contrastive = criterion(output1, output2, label)
            loss_contrastive.backward()
            optimizer.step()
            if iteration_number % 10 == 0:
                print("epoch_number number {}\n iteration_number number {}\n Current loss {}\n".format(epoch,
                                                                                                       iteration_number,
                                                                                                       loss_contrastive.item()))
                iteration_number += 10
                counter.append(iteration_number)
                loss_history.append(loss_contrastive.item())

    plt.plot(counter, loss_history)
    plt.show()
    # torch.save(net, 'model.pkl')
